// pre-commit makes sure the changed code is formatted and linted.
//
// Rationale:
// - Fixing formatting and linting errors are always fast and easy to fix.
// - This prevents over-complicated merge conflicts
// - This prevents small formatting/linting fix commits
"use strict";
const helpers_1 = require("../helpers/helpers");
const tsfmt = require("typescript-formatter");
const fs_1 = require("fs");
const path_1 = require("path");
const glob = require("glob");
const tslint_1 = require("tslint");
const util_1 = require("../../src/util");
/* tslint:disable:no-console */
let allTsFiles = () => {
    let tsConfig = JSON.parse(fs_1.readFileSync(path_1.join(process.cwd(), 'tsconfig.json'), 'UTF-8'));
    let globs = tsConfig && tsConfig.include ? tsConfig.include : ['src/**/*.ts', 'test/**/*.ts'];
    let files = [];
    globs.forEach(g => glob.sync(g).forEach(file => files.push(file)));
    return files;
};
let changed = helpers_1.findChangedFiles();
let files = changed.length === 0 ? allTsFiles() : helpers_1.filterTsFiles(changed);
let lintFiles = () => {
    let linter = new tslint_1.Linter({ fix: false, formatter: 'prose' });
    let configurationFile = tslint_1.Linter.loadConfigurationFromPath(process.cwd() + '/tslint.json');
    let configuration = {
        rules: configurationFile.rules,
        rulesDirectory: configurationFile.rulesDirectory
    };
    let success = true;
    files.forEach((fileName) => {
        let contents = fs_1.readFileSync(fileName, 'utf8');
        linter.lint(fileName, contents, configuration);
        let results = linter.getResult();
        results.failures.forEach((failure) => {
            success = false;
            let line = failure.getStartPosition().getLineAndCharacter().line + 1;
            let column = failure.getStartPosition().getLineAndCharacter().character;
            console.log(`${util_1.absolutePath(fileName)}:${line}:${column} ${failure.getFailure()}`);
        });
    });
    return success;
};
tsfmt.processFiles(files, {
    verify: true, replace: false, verbose: false, baseDir: process.cwd(), editorconfig: true, tslint: true, tsfmt: true, tsconfig: true
}).then((resultList) => {
    let unformattedFiles = [];
    Object.keys(resultList).forEach((key) => {
        let result = resultList[key];
        if (result.error) {
            unformattedFiles.push(result.fileName);
        }
    });
    if (unformattedFiles.length === 0) {
        console.log(`All ${files.length} files were formatted`);
        if (lintFiles()) {
            console.log(`All ${files.length} files were linted`);
            process.exit(0);
        }
        else {
            console.log(`There were linting errors`);
            process.exit(1);
        }
    }
    else {
        console.error(`The following files were not formatted:\r\n  ${unformattedFiles.join('\r\n  ')}`);
        process.exit(1);
    }
});
//# sourceMappingURL=pre-commit.js.map